/*
 * Copyright (C) 2011-2012 Texas Instruments Inc.
 *
 * Author: Brijesh Singh <bksingh@ti.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "gstomx_noisefilter.h"
#include "gstomx.h"

GSTOMX_BOILERPLATE (GstOmxNoiseFilter, gst_omx_noisefilter, GstOmxBaseVfpc, GST_OMX_BASE_VFPC_TYPE);

static GstStaticPadTemplate sink_template =
        GST_STATIC_PAD_TEMPLATE ("sink",
                GST_PAD_SINK,
                GST_PAD_ALWAYS,
                GST_STATIC_CAPS (GST_VIDEO_CAPS_YUV_STRIDED (
                        "{YUY2}", "[ 0, max ]"))
        );

static GstStaticPadTemplate src_template =
        GST_STATIC_PAD_TEMPLATE ("src",
                GST_PAD_SRC,
                GST_PAD_ALWAYS,
                GST_STATIC_CAPS (GST_VIDEO_CAPS_YUV ( "{NV12}" ))
        );

static void
type_base_init (gpointer g_class)
{
    GstElementClass *element_class;

    element_class = GST_ELEMENT_CLASS (g_class);

    {
        GstElementDetails details;

        details.longname = "OpenMAX IL for OMX.TI.VPSSM3.VFPC.NF component";
        details.klass = "Filter";
        details.description = "Noise filter using VPSS Noise filter module";
        details.author = "Brijesh Singh";

        gst_element_class_set_details (element_class, &details);
    }

    gst_element_class_add_pad_template (element_class,
        gst_static_pad_template_get (&sink_template));

    gst_element_class_add_pad_template (element_class,
        gst_static_pad_template_get (&src_template));
}

static void
type_class_init (gpointer g_class,
                 gpointer class_data)
{
}

static GstCaps*
create_src_caps (GstOmxBaseFilter *omx_base)
{
    GstCaps *caps;
    GstOmxBaseVfpc *self;
    GstStructure *struc;
    int width, height;
    self = GST_OMX_BASE_VFPC (omx_base);
    caps = gst_pad_peer_get_caps (omx_base->srcpad);


    if (NULL == caps || gst_caps_is_empty (caps))
    {
        width = self->in_width;
        height = self->in_height;
    }
    else
    {
        GstStructure *s;

        s = gst_caps_get_structure (caps, 0);

        if (!(gst_structure_get_int (s, "width", &width) &&
            gst_structure_get_int (s, "height", &height)))
        {
            width = self->in_width;
            height = self->in_height;
        }
    }

    caps = gst_caps_new_empty ();
    struc = gst_structure_new (("video/x-raw-yuv"),
            "width",  G_TYPE_INT, width,
            "height", G_TYPE_INT, height,
            "format", GST_TYPE_FOURCC, GST_MAKE_FOURCC ('N', 'V', '1', '2'),
            NULL);

    if (self->framerate_denom)
    {
        gst_structure_set (struc,
        "framerate", GST_TYPE_FRACTION, self->framerate_num, self->framerate_denom, NULL);
    }

    gst_caps_append_structure (caps, struc);

    return caps;
}

static void
omx_setup (GstOmxBaseFilter *omx_base)
{
    GOmxCore *gomx;
    OMX_ERRORTYPE err;
    OMX_PARAM_PORTDEFINITIONTYPE paramPort;
    OMX_PARAM_BUFFER_MEMORYTYPE memTypeCfg;
    OMX_PARAM_VFPC_NUMCHANNELPERHANDLE numChannels;
    OMX_CONFIG_VIDCHANNEL_RESOLUTION chResolution;
    OMX_CONFIG_ALG_ENABLE algEnable;
    GstOmxBaseVfpc *self;

    gomx = (GOmxCore *) omx_base->gomx;
    self = GST_OMX_BASE_VFPC (omx_base);

    GST_LOG_OBJECT (self, "begin");

    /* Set the output caps */
    gst_pad_set_caps (omx_base->srcpad, create_src_caps (omx_base));

    /* Setting Memory type at input port to Raw Memory */
    GST_LOG_OBJECT (self, "Setting input port to Raw memory");

    _G_OMX_INIT_PARAM (&memTypeCfg);
    memTypeCfg.nPortIndex = self->input_port_index;
    memTypeCfg.eBufMemoryType = OMX_BUFFER_MEMORY_DEFAULT;
    err = OMX_SetParameter (gomx->omx_handle, OMX_TI_IndexParamBuffMemType, &memTypeCfg);

    if (err != OMX_ErrorNone)
      return;

    /* Setting Memory type at output port to Raw Memory */
    GST_LOG_OBJECT (self, "Setting output port to Raw memory");

    _G_OMX_INIT_PARAM (&memTypeCfg);
    memTypeCfg.nPortIndex = self->output_port_index;
    memTypeCfg.eBufMemoryType = OMX_BUFFER_MEMORY_DEFAULT;
    err = OMX_SetParameter (gomx->omx_handle, OMX_TI_IndexParamBuffMemType, &memTypeCfg);

    if (err != OMX_ErrorNone)
      return;

    /* Input port configuration. */
    GST_INFO_OBJECT (self, "Setting port definition (input)");

    G_OMX_PORT_GET_DEFINITION (omx_base->in_port, &paramPort);
    paramPort.format.video.nFrameWidth = (self->in_width+31) & 0xFFFFFFE0;
    paramPort.format.video.nFrameHeight = (self->in_height+31) & 0xFFFFFFE0;
    paramPort.format.video.nStride = ((self->in_width+31) & 0xFFFFFFE0)*2;
    paramPort.format.video.eCompressionFormat = OMX_VIDEO_CodingUnused;
    paramPort.format.video.eColorFormat = OMX_COLOR_FormatYCbYCr;
    paramPort.nBufferSize = (paramPort.format.video.nStride * paramPort.format.video.nFrameHeight * 2);
    paramPort.nBufferAlignment = 0;
    paramPort.bBuffersContiguous = 0;
    G_OMX_PORT_SET_DEFINITION (omx_base->in_port, &paramPort);
    g_omx_port_setup (omx_base->in_port, &paramPort);

    /* Output port configuration. */
    GST_INFO_OBJECT (self, "Setting port definition (output)");

    G_OMX_PORT_GET_DEFINITION (omx_base->out_port, &paramPort);
    paramPort.format.video.nFrameWidth = (self->in_width+31) & 0xFFFFFFE0;
    paramPort.format.video.nFrameHeight = (self->in_height+31) & 0xFFFFFFE0;
    paramPort.format.video.nStride = (self->in_width+31) & 0xFFFFFFE0;
    paramPort.format.video.eCompressionFormat = OMX_VIDEO_CodingUnused;
    paramPort.format.video.eColorFormat = OMX_COLOR_FormatYUV420SemiPlanar;
    paramPort.nBufferSize = (paramPort.format.video.nStride * paramPort.format.video.nFrameHeight * 3) >> 1;
    paramPort.nBufferAlignment = 0;
    paramPort.bBuffersContiguous = 0;
    G_OMX_PORT_SET_DEFINITION (omx_base->out_port, &paramPort);
    g_omx_port_setup (omx_base->out_port, &paramPort);

    /* Set number of channels */
    GST_INFO_OBJECT (self, "Setting number of channels");

    _G_OMX_INIT_PARAM (&numChannels);
    numChannels.nNumChannelsPerHandle = 16;
    err = OMX_SetParameter (gomx->omx_handle,
        (OMX_INDEXTYPE) OMX_TI_IndexParamVFPCNumChPerHandle, &numChannels);

    if (err != OMX_ErrorNone)
        return;

    int i = 0;

    for( i = 0; i < numChannels.nNumChannelsPerHandle; ++i) {

      /* Set input channel resolution */
      GST_INFO_OBJECT (self, "Setting input resolution for channel #%d", i);

      _G_OMX_INIT_PARAM (&chResolution);
      chResolution.Frm0Width = (self->in_width+31) & 0xFFFFFFE0;
      chResolution.Frm0Height = (self->in_height+31) & 0xFFFFFFE0;
      chResolution.Frm0Pitch = ((self->in_width+31) & 0xFFFFFFE0)*2;
      chResolution.Frm1Width = chResolution.Frm0Width; //0;
      chResolution.Frm1Height = chResolution.Frm0Height; //0;
      chResolution.Frm1Pitch = chResolution.Frm0Pitch; //0;
      chResolution.FrmStartX = 0;
      chResolution.FrmStartY = 0;
      chResolution.FrmCropWidth = self->in_width;
      chResolution.FrmCropHeight = self->in_height;
      chResolution.eDir = OMX_DirInput;
      chResolution.nChId = i;
      err = OMX_SetConfig (gomx->omx_handle, OMX_TI_IndexConfigVidChResolution, &chResolution);

      if (err != OMX_ErrorNone)
        return;

      /* Set output channel resolution */
      GST_INFO_OBJECT (self, "Setting output resolution for channel #%d", i);

      _G_OMX_INIT_PARAM (&chResolution);
      chResolution.Frm0Width = (self->in_width+31) & 0xFFFFFFE0;
      chResolution.Frm0Height = (self->in_height+31) & 0xFFFFFFE0;
      chResolution.Frm0Pitch = ((self->in_width+31) & 0xFFFFFFE0);
      chResolution.Frm1Width = chResolution.Frm0Width; //0;
      chResolution.Frm1Height = chResolution.Frm0Height; //0;
      chResolution.Frm1Pitch = chResolution.Frm0Pitch; //0;
      chResolution.FrmStartX = 0;
      chResolution.FrmStartY = 0;
      chResolution.FrmCropWidth = self->in_width;
      chResolution.FrmCropHeight = self->in_height;
      chResolution.eDir = OMX_DirOutput;
      chResolution.nChId = i;
      err = OMX_SetConfig (gomx->omx_handle, OMX_TI_IndexConfigVidChResolution, &chResolution);

      if (err != OMX_ErrorNone)
        return;

      /* int j = 0; */
      /* for( j = 0; j < 2; ++j ) */
      {

	/* OMX_PARAM_VIDEONOISEFILTERTYPE filterType; */
	/* _G_OMX_INIT_PARAM (&filterType); */
	/* filterType.nPortIndex = j; */
	/* filterType.eMode = OMX_VideoNoiseFilterModeOff; */
	/* err = OMX_SetConfig (gomx->omx_handle, (OMX_INDEXTYPE) OMX_IndexParamVideoNoiseFilter, &filterType); */

	/* if (err != OMX_ErrorNone) { */
	/*   fprintf(stderr, "Could not set Filter Type for index %d\n", j); */
	/* } */

	_G_OMX_INIT_PARAM (&algEnable);
	algEnable.nPortIndex = self->input_port_index;
	algEnable.nChId = i;
	algEnable.bAlgBypass = 1;

	err = OMX_SetConfig (gomx->omx_handle, (OMX_INDEXTYPE) OMX_TI_IndexConfigAlgEnable, &algEnable);

	if (err != OMX_ErrorNone)
	  return;

	_G_OMX_INIT_PARAM (&algEnable);
	algEnable.nPortIndex = self->output_port_index;
	algEnable.nChId = i;
	algEnable.bAlgBypass = 1;

	err = OMX_SetConfig (gomx->omx_handle, (OMX_INDEXTYPE) OMX_TI_IndexConfigAlgEnable, &algEnable);

	if (err != OMX_ErrorNone)
	  return;
      }
    }
}

static void
type_instance_init (GTypeInstance *instance,
                    gpointer g_class)
{
    GstOmxBaseVfpc *self;

    self = GST_OMX_BASE_VFPC (instance);

    self->omx_setup = omx_setup;
    g_object_set (self, "port-index", 0, NULL);
}
